import os
import gmsh
import time

from fiqus.utils.Utils import GmshUtils
from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.data import DataFiQuS as dF
from fiqus.geom_generators.GeometryMultipole import Geometry
from fiqus.mesh_generators.MeshMultipole import Mesh
from fiqus.getdp_runners.RunGetdpMultipole import RunGetdpMultipole
from fiqus.getdp_runners.RunGetdpMultipole import AssignNaming
from fiqus.post_processors.PostProcessMultipole import PostProcess
from fiqus.plotters.PlotPythonMultipole import PlotPythonMultipole


class MainMultipole:
    def __init__(self, fdm: dF.FDM = None, sdm: dF.FiQuSSettings = None, rgd_path: str = None,
                 verbose: bool = None):
        """
        Main class for working with simulations for multipole type magnets
        :param fdm: FiQuS data model
        :param rgd_path: ROXIE geometry data path
        :param sdm: settings data model
        :param verbose: if True, more info is printed in the console
        """
        self.fdm = fdm
        self.sett = sdm
        self.rgd = rgd_path
        self.verbose = verbose

        self.settings = None
        self.geom_folder = None
        self.mesh_folder = None
        self.solution_folder = None
        self.model_file = None

    def generate_geometry(self, gui: bool = False):
        geom = Util.read_data_from_yaml(self.rgd, dF.FiQuSGeometry)
        gg = Geometry(data=self.fdm, geom=geom, sett=self.sett, geom_folder=self.geom_folder, verbose=self.verbose)
        self.model_file = gg.model_file
        gg.saveStrandPositions()
        if self.fdm.magnet.geometry.with_iron_yoke:
            gg.constructIronGeometry()
        gg.constructWedgeGeometry()
        gg.constructCoilGeometry()
        gg.buildDomains()
        gg.saveBoundaryRepresentationFile()
        gg.loadBoundaryRepresentationFile()
        gg.updateTags()
        gg.saveAuxiliaryFile()
        gg.ending_step(gui)

    def load_geometry(self, gui: bool = False):
        gu = GmshUtils(self.geom_folder, self.verbose)
        gu.initialize()
        self.model_file = f"{os.path.join(self.geom_folder, self.fdm.general.magnet_name)}.brep"
        gmsh.option.setString('Geometry.OCCTargetUnit', 'M')  # set units to meters
        gmsh.open(self.model_file)
        if gui:
            gu.launch_interactive_GUI()

    def pre_process(self, gui: bool = False):
        pass

    def mesh(self, gui: bool = False):
        mm = Mesh(data=self.fdm, sett=self.sett, mesh_folder=self.mesh_folder, verbose=self.verbose)
        self.model_file = mm.model_file
        mm.loadAuxiliaryFile()
        mm.getIronCurvesTags()
        mm.defineMesh()
        mm.fragment()
        mm.createPhysicalGroups()
        mm.updateAuxiliaryFile()
        mm.assignRegionsTags()
        mm.saveRegionFile()
        mm.setMeshOptions()
        mm.generateMesh()
        mm.checkMeshQuality()
        mm.saveMeshFile()
        mm.ending_step(gui)
        return mm.mesh_parameters

    def load_mesh(self, gui: bool = False):
        gu = GmshUtils(self.geom_folder, self.verbose)
        gu.initialize()
        self.model_file = f"{os.path.join(self.mesh_folder, self.fdm.general.magnet_name)}.msh"
        gmsh.open(self.model_file)
        if gui:
            gu.launch_interactive_GUI()

    def solve_and_postprocess_getdp(self, gui: bool = False):
        an = AssignNaming(data=self.fdm)
        rg = RunGetdpMultipole(data=an, sett=self.sett, solution_folder=self.solution_folder, settings=self.settings,
                               verbose=self.verbose)
        self.model_file = rg.model_file
        rg.loadRegionFile()
        rg.assemblePro()
        start_time = time.time()
        rg.solve_and_postprocess()
        rg.ending_step(gui)
        return time.time() - start_time

    def post_process_getdp(self, gui: bool = False):
        an = AssignNaming(data=self.fdm)
        rg = RunGetdpMultipole(data=an, sett=self.sett, solution_folder=self.solution_folder, settings=self.settings,
                               verbose=self.verbose)
        self.model_file = rg.model_file
        rg.loadRegionFile()
        rg.assemblePro()
        rg.postprocess()
        rg.ending_step(gui)

    def post_process_python(self, gui: bool = False):
        if self.fdm.run.type == 'post_process_python_only':
            an = AssignNaming(data=self.fdm)
            data = an.data
        else:
            data = self.fdm
        pp = PostProcess(data=data, sett=self.sett, solution_folder=self.solution_folder, verbose=self.verbose)
        self.model_file = pp.model_file
        pp.loadStrandPositions()
        if self.fdm.magnet.postproc.plot_all:
            pp.loadHalfTurnCornerPositions()
        pp.postProcess()
        pp.ending_step(gui)
        return pp.postprocess_parameters

    def plot_python(self):
        os.chdir(self.solution_folder)
        p = PlotPythonMultipole(self.fdm)
        p.dummy_plot_func()


import os
import getpass

from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.utils.Utils import CheckForExceptions as Check
from fiqus.data import DataFiQuS as dF
from fiqus.mains.MainMultipole import MainMultipole


class MainFiQuS:
    def __init__(self, input_file_path: str = None, model_folder: str = None, verbose: bool = True, fdm=None):
        """
        Main class for working with FiQuS simulations
        :param input_file_path: input file name
        :param verbose: if True, more info is printed in the console
        """
        self.start_folder = os.getcwd()
        self.wrk_folder = model_folder
        self.verbose = verbose
        if self.verbose:
            Util.print_welcome_graphics()

        # Load yaml input file
        if not fdm:
            self.fdm = Util.read_data_from_yaml(input_file_path, dF.FDM)
        else:
            self.fdm = fdm

        # Check for input errors
        Check.check_inputs(self.fdm.run)

        # Initialize Main object
        if self.fdm.magnet.type == 'CCT_straight':
            self.main_magnet = MainCCT(fdm=self.fdm, verbose=verbose)

        elif self.fdm.magnet.type == 'multipole':
            # Load settings
            self.sdm = Util.read_data_from_yaml(f'{input_file_path[:-5]}.set', dF.FiQuSSettings)

            self.main_magnet = MainMultipole(fdm=self.fdm, sdm=self.sdm, rgd_path=f'{input_file_path[:-5]}.geom',
                                             verbose=verbose)

        else:
            raise ValueError(f'FiQuS does not support magnet type: {self.fdm.magnet.type}!')

        # Load user paths for executables and additional files
        user_name = getpass.getuser()
        if verbose:
            print(f'FiQuS is running on machine with user name: {user_name}')
        if user_name in ['root', 'MP-WIN-02$']:
            user_name = 'SYSTEM'
        path_to_settings_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), "tests", f"settings.{user_name}.yaml")
        if verbose:
            print(f'FiQuS is using settings file: {path_to_settings_file}')
        self.main_magnet.settings = Util.read_data_from_yaml(path_to_settings_file, dict)

        # Save Model/Geometry/Mesh/Solution folder paths
        Util.prep_folder(self.wrk_folder)
        self.save_folders()

        # Build magnet
        self.summary = dict.fromkeys(['SJ', 'SICN', 'SIGE', 'Gamma', 'nodes', 'solution_time', 'overall_error', 'minimum_diff', 'maximum_diff'])
        self.build_magnet()

    def save_folders(self):
        def _check_and_generate_path(folder_type: str = None):
            if folder_type == 'Geometry':
                folder = self.wrk_folder
                ref_nr = self.fdm.run.geometry
            elif folder_type == 'Mesh':
                folder = self.main_magnet.geom_folder
                ref_nr = self.fdm.run.mesh
            elif folder_type == 'Solution':
                folder = self.main_magnet.mesh_folder
                ref_nr = self.fdm.run.solution
            else:
                raise Exception('Incompatible type.')

            required_folder = folder_type in required_folders
            if self.fdm.run.overwrite and folder_type == (required_folders[0] if required_folders else None):
                Check.check_overwrite_conditions(folder_type=folder_type, folder=folder, ref_nr=ref_nr)
            return Util.get_folder_path(folder_type=folder_type, folder=folder, ref_nr=ref_nr,
                                        overwrite=self.fdm.run.overwrite, required_folder=required_folder)

        if self.fdm.run.type == 'start_from_yaml':
            required_folders = ['Geometry', 'Mesh', 'Solution']
        elif self.fdm.run.type == 'geometry_only':
            required_folders = [] if self.fdm.run.geometry and not self.fdm.run.overwrite else ['Geometry']
        elif self.fdm.run.type == 'mesh_and_solve_with_post_process_python':
            required_folders = ['Mesh', 'Solution']
        elif self.fdm.run.type == 'mesh_only':
            required_folders = [] if self.fdm.run.mesh and not self.fdm.run.overwrite else ['Mesh']
        elif self.fdm.run.type in ['solve_with_post_process_python', 'solve_only']:
            required_folders = ['Solution']
        else:  # post_process_getdp_only or post_process_python_only or plot_python
            required_folders = []

        fdm = self.main_magnet.fdm.magnet
        self.main_magnet.geom_folder = _check_and_generate_path(folder_type='Geometry')
        Util.write_data_to_yaml(os.path.join(self.main_magnet.geom_folder, 'geometry.yaml'), fdm.geometry.dict())
        if not self.fdm.run.type == 'geometry_only':
            self.main_magnet.mesh_folder = _check_and_generate_path(folder_type='Mesh')
            Util.write_data_to_yaml(os.path.join(self.main_magnet.mesh_folder, 'mesh.yaml'), fdm.mesh.dict())
        if not (self.fdm.run.type == 'geometry_only' or self.fdm.run.type == 'mesh_only'):
            self.main_magnet.solution_folder = _check_and_generate_path(folder_type='Solution')
            Util.write_data_to_yaml(os.path.join(self.main_magnet.solution_folder, 'solve.yaml'), fdm.solve.dict())
            Util.write_data_to_yaml(os.path.join(self.main_magnet.solution_folder, 'postproc.yaml'), fdm.postproc.dict())

    def build_magnet(self):
        if self.fdm.run.type == 'start_from_yaml':  # needs 3 files (yaml, set, geom)
            self.main_magnet.generate_geometry()
            self.main_magnet.pre_process()
            self.main_magnet.load_geometry()
            for key, value in self.main_magnet.mesh().items():
                self.summary[key] = value
            self.summary['solution_time'] = self.main_magnet.solve_and_postprocess_getdp()
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == 'geometry_only':
            if len(os.listdir(self.main_magnet.geom_folder)) == 1:
                self.main_magnet.generate_geometry()  # needs 3 files (yaml, set, geom)
                self.main_magnet.pre_process(gui=self.main_magnet.fdm.run.launch_gui)
            else:
                self.main_magnet.load_geometry(gui=self.main_magnet.fdm.run.launch_gui)  # needs 2 files (yaml, brep)
        elif self.fdm.run.type == 'mesh_and_solve_with_post_process_python':  # needs 5 files (yaml, strs/map2d, set, brep, aux)
            self.main_magnet.load_geometry()
            for key, value in self.main_magnet.mesh().items():
                self.summary[key] = value
            self.summary['solution_time'] = self.main_magnet.solve_and_postprocess_getdp()
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == 'mesh_only':
            if len(os.listdir(self.main_magnet.mesh_folder)) == 1:
                self.main_magnet.load_geometry()  # needs 3 files (yaml, brep, aux)
                for key, value in self.main_magnet.mesh(gui=self.main_magnet.fdm.run.launch_gui).items():
                    self.summary[key] = value
            else:
                self.main_magnet.load_mesh(gui=self.main_magnet.fdm.run.launch_gui)  # needs 2 files (yaml, msh)
        elif self.fdm.run.type == 'solve_with_post_process_python':  # needs 5 files (yaml, strs/map2d, set, msh, reg)
            self.summary['solution_time'] = self.main_magnet.solve_and_postprocess_getdp()
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == 'solve_only':  # needs 5 files (yaml, strs/map2d, set, msh, reg)
            self.summary['solution_time'] = self.main_magnet.solve_and_postprocess_getdp(gui=self.main_magnet.fdm.run.launch_gui)
        elif self.fdm.run.type == 'post_process_getdp_only':
            self.main_magnet.post_process_getdp(gui=self.main_magnet.fdm.run.launch_gui)
        elif self.fdm.run.type == 'post_process_python_only':
            for key, value in self.main_magnet.post_process_python(gui=self.main_magnet.fdm.run.launch_gui).items():
                self.summary[key] = value
        elif self.fdm.run.type == 'plot_python':
            self.main_magnet.plot_python()
        os.chdir(self.start_folder)

if __name__ == "__main__":
    # if len(sys.argv) < 4:
    #     mp = MainFiQuS(sys.argv[1], sys.argv[2])

    magnet = 'MQXA'
    yaml_file = f'{magnet}.yaml'
    output_folder = os.getcwd()
    sim_result = MainFiQuS(input_file_path=yaml_file, model_folder=os.path.join(output_folder, magnet))



from pydantic import BaseModel
from typing import (List, Literal)


class Threshold(BaseModel):
    """
        Level 2: Class for FiQuS Multipole
    """
    SizeMin: float = None
    SizeMax: float = None
    DistMin: float = None
    DistMax: float = None


class GeometryMultipole(BaseModel):
    """
        Level 2: Class for FiQuS Multipole
    """
    with_iron_yoke: bool = None


class MeshMultipole(BaseModel):
    """
        Level 2: Class for FiQuS Multipole
    """
    default_mesh: bool = None
    mesh_iron: Threshold = Threshold()
    mesh_coil: Threshold = Threshold()
    MeshSizeMin: float = None  # sets gmsh Mesh.MeshSizeMin
    MeshSizeMax: float = None  # sets gmsh Mesh.MeshSizeMax
    MeshSizeFromCurvature: float = None  # sets gmsh Mesh.MeshSizeFromCurvature
    Algorithm: int = None  # sets gmsh Mesh.Algorithm
    AngleToleranceFacetOverlap: float = None  # sets gmsh Mesh.AngleToleranceFacetOverlap
    ElementOrder: int = None  # sets gmsh Mesh.ElementOrder
    Optimize: int = None  # sets gmsh Mesh.Optimize


class SolveMultipoleFiQuS(BaseModel):
    """
        Level 2: Class for FiQuS Multipole
    """
    I_initial: List[float] = None
    pro_template: str = None  # file name of .pro template file


class PostProcMultipole(BaseModel):
    """
        Level 2: Class for FiQuS Multipole
    """
    compare_to_ROXIE: str = None
    plot_all: str = None
    variables: List[str] = None  # Name of variables to post-process, like "b" for magnetic flux density
    volumes: List[str] = None  # Name of domains to post-process, like "powered"
    file_exts: List[str] = None  # Name of file extensions to output to, like "pos"
    additional_outputs: List[str] = None  # Name of software specific input files to prepare, like "LEDET3D"


class MPDM(BaseModel):
    """
        Level 2: Class for FiQuS Multipole
    """
    type: Literal['multipole']
    geometry: GeometryMultipole = GeometryMultipole()
    mesh: MeshMultipole = MeshMultipole()
    solve: SolveMultipoleFiQuS = SolveMultipoleFiQuS()
    postproc: PostProcMultipole = PostProcMultipole()



from pydantic import BaseModel
from typing import (Dict, List, Union, Literal)
from fiqus.data.DataRoxieParser import RoxieData
from fiqus.data.DataFiQuSMultipole import MPDM


class MonoFiQuS(BaseModel):
    """
        Rutherford cable type
    """
    type: Literal['Mono']
    bare_cable_width: float = None
    bare_cable_height_mean: float = None


class RibbonFiQuS(BaseModel):
    """
        Rutherford cable type
    """
    type: Literal['Ribbon']
    bare_cable_width: float = None
    bare_cable_height_mean: float = None


class RutherfordFiQuS(BaseModel):
    """
        Rutherford cable type
    """
    type: Literal['Rutherford']
    bare_cable_width: float = None
    bare_cable_height_mean: float = None


class ConductorFiQuS(BaseModel):
    """
        Class for conductor type
    """
    cable: Union[RutherfordFiQuS, RibbonFiQuS, MonoFiQuS] = {'type': 'Rutherford'}


class GeneralSetting(BaseModel):
    """
        Class for general information on the case study
    """
    I_ref: List[float] = None


class ModelDataSetting(BaseModel):
    """
        Class for model data
    """
    general_parameters: GeneralSetting = GeneralSetting()
    conductors: Dict[str, ConductorFiQuS] = {}

#######################################################################################################################


class FiQuSGeometry(BaseModel):
    """
        Class for Roxie data
    """
    Roxie_Data: RoxieData = RoxieData()


class FiQuSSettings(BaseModel):
    """
        Class for FiQuS model
    """
    Model_Data_GS: ModelDataSetting = ModelDataSetting()


class RunFiQuS(BaseModel):
    """
        Class for FiQuS run
    """
    type: str = None
    geometry: int = None
    mesh: int = None
    solution: int = None
    launch_gui: bool = None
    overwrite: bool = None


class GeneralFiQuS(BaseModel):
    """
        Class for FiQuS general
    """
    magnet_name: str = None


class FDM(BaseModel):
    """
        Class for FiQuS
    """
    general: GeneralFiQuS = GeneralFiQuS()
    run: RunFiQuS = RunFiQuS()
    magnet: Union[MPDM] = {'type': 'multipole'}



from pydantic import BaseModel
from typing import (Dict, List)


class Coord(BaseModel):
    x: float = None
    y: float = None


class Area(BaseModel):
    loop: int = None
    surface: int = None


class Region(BaseModel):
    points: Dict[str, int] = {}
    lines: Dict[str, int] = {}
    areas: Dict[str, Area] = {}


class BlockData(BaseModel):
    half_turns: Region = Region()
    current_sign: int = None


class Block(BaseModel):
    blocks: Dict[int, BlockData] = {}
    conductor_name: str = None
    conductors_number: int = None


class Winding(BaseModel):
    windings: Dict[int, Block] = {}


class Layer(BaseModel):
    layers: Dict[int, Winding] = {}


class Pole(BaseModel):
    poles: Dict[int, Layer] = {}
    bore_center: Coord = Coord()


class Iron(BaseModel):
    quadrants: Dict[int, Region] = {}
    max_radius: float = 0.


class Order(BaseModel):
    coil: int = None
    pole: int = None
    layer: int = None
    winding: int = None
    block: int = None


class Coil(BaseModel):
    coils: Dict[int, Pole] = {}
    electrical_order: List[Order] = []


class Geometry(BaseModel):
    coil: Coil = Coil()
    iron: Iron = Iron()
    wedges: Region = Region()
    air: Region = Region()
    air_inf: Region = Region()


class GroupType(BaseModel):
    curves: Dict[str, int] = {}
    surfaces: Dict[str, int] = {}
    volumes: Dict[str, int] = {}


class Domain(BaseModel):
    groups_surfaces: Dict[str, List[int]] = {}
    physical_groups: GroupType = GroupType()


class MultipoleData(BaseModel):
    geometries: Geometry = Geometry()
    domains: Domain = Domain()


# if __name__ == "__main__":
#     write = True
#     read = False
#
#     def read_regions(regions_file_name):
#         with open(regions_file_name, 'r') as stream:
#             yaml_str = ruamel.yaml.safe_load(stream)
#         return MultipoleData(**yaml_str)
#
#     if write:
#         model = MultipoleData()
#         with open('FiQuS_data.yaml', 'w') as yaml_file:
#             ruamel.yaml.dump(model.dict(), yaml_file, default_flow_style=False)



import re
import inspect
import pandas as pd


class ParserPOS:

    def __init__(self, pos_file_path):
        """
        Read pos file and returns its content as object attribute .data_dict that is a dictionary.
        :param pos_file_path: Full path to .pos file, including file name and extension.
        """
        self._mesh_format_markers = {'s': '$MeshFormat', 'e': '$EndMeshFormat'}
        self._nodes_markers = {'s': '$Nodes', 'e': '$EndNodes'}
        self._elements_markers = {'s': '$Elements', 'e': '$EndElements'}
        self._elements_node_data_markers = {'s': '$ElementNodeData', 'e': '$EndElementNodeData'}
        self._physical_name_markers = {'s': 'PhysicalNames', 'e': '$EndPhysicalNames'}

        with open(pos_file_path) as f:
            self._contents = f.read()
        # node properteis
        self._node_numbers = []
        self._node_coordinates = []
        # element properties
        self._element_numbers = []
        self._element_types = []
        self._element_number_of_tags = []
        self._element_physical_tags = []
        self._element_elementary_tags = []
        self._element_node_numbers = []
        # elements_node_data properties
        self._data_element_tags = []
        self._data_num_nodes_per_element = []
        self._values = []
        # parse the content (output is in this class attributes)
        self._parse()

    def __get_content(self, markers_dict):
        """
        Gets text string between two markers specified in markers_dict
        """
        return self._contents[self._contents.find(markers_dict['s']) + len(markers_dict['s']):self._contents.find(markers_dict['e'])]

    @staticmethod
    def __get_lines(data_str, empty_lines=[0, -1]):
        """
        Converts text string into a list of lines
        """
        data_str = re.sub('\n', "'", data_str)
        data_str = re.sub('"', '', data_str)
        str_list = re.split("'", data_str)
        for empty_line in empty_lines:
            if not str_list.pop(empty_line) == '':
                raise ValueError('Error in parsing lines')
        return str_list

    @staticmethod
    def __list_from_list_at_pos(dest_list, source_list, d_type, pos):
        """
        This one is a bit complex. It converts a supplied in source list and appends it destination list.
        It uses data type and position to append to the right type to the destination list from the right place in the source list.
        Position could be an integer, a list of integers or a tuple with a reference variables to figure out positions from them.
        """
        if isinstance(pos, int):
            dest_list.append(d_type(source_list[pos]))
        elif isinstance(pos, list):
            dest_list.append([d_type(source_list[p]) for p in pos])
        elif isinstance(pos, tuple):
            offset = pos[0]
            dimension = pos[1]
            ref_list = pos[2]
            current_num_elem = ref_list[len(dest_list)]  # get current element
            pos = list(range(offset, dimension * current_num_elem + offset))
            dest_list.append([d_type(source_list[p]) for p in pos])

    def __parse_lines(self, lines, list_of_outputs_lists, list_of_data_types, list_of_positions, len_check, ):
        """
        Simply loop through lines and append to class attributes lists. Basic error check if number of lines in the list matches that declared in the file.
        """
        for line in lines:
            values = re.split(' ', line)
            for dest_list, d_type, pos in zip(list_of_outputs_lists, list_of_data_types, list_of_positions):
                self.__list_from_list_at_pos(dest_list, values, d_type, pos)
        for out_list in list_of_outputs_lists:
            if not len_check == len(out_list):
                raise ValueError(f'Error in parsing {inspect.stack()[1].function}')

    def _mesh_format(self):
        """
        Parse mesh_generators field and assign it to the class attribute
        """
        self.mesh_format = self.__get_content(self._mesh_format_markers)

    def _physical_names(self):
        """
        Parse physical_names field and assign it to the class attribute
        """
        self.physical_names = self.__get_content(self._physical_name_markers)

    def _nodes(self):
        """
        Parse nodes  and assign it to the class attributes
        """
        data_str = self.__get_content(self._nodes_markers)
        lines = self.__get_lines(data_str)
        self._number_of_nodes = int(lines.pop(0))
        list_of_outputs_lists = [self._node_numbers, self._node_coordinates]
        list_of_data_types = [int, float]
        list_of_positions = [0, [1, 2, 3]]
        self.__parse_lines(lines, list_of_outputs_lists, list_of_data_types, list_of_positions, self._number_of_nodes)

    def _elements(self):
        """
        Parse elements  and assign it to the class attributes
        """
        data_str = self.__get_content(self._elements_markers)
        lines = self.__get_lines(data_str)
        self._number_of_elements = int(lines.pop(0))
        list_of_outputs_lists = [self._element_numbers, self._element_types, self._element_number_of_tags, self._element_physical_tags, self._element_elementary_tags, self._element_node_numbers]
        list_of_data_types = [int, int, int, int, int, int]
        list_of_positions = [0, 1, 2, 3, 4, (5, 1, self._data_num_nodes_per_element)]
        self.__parse_lines(lines, list_of_outputs_lists, list_of_data_types, list_of_positions, self._number_of_elements)

    def _elements_node_data(self):
        """
        Parse elements data and assign it to the class attributes
        """
        data_str = self.__get_content(self._elements_node_data_markers)
        lines = self.__get_lines(data_str)
        self._numStringTags = int(lines.pop(0))
        self._stringTags = str(lines.pop(0))
        self._numRealTags = int(lines.pop(0))
        self._realTags = float(lines.pop(0))
        self._numIntegerTags = int(lines.pop(0))
        self._integerTags = int(lines.pop(0))
        self._entityDim = int(lines.pop(0))
        self._number_of_elements_data = int(lines.pop(0))
        self._not_sure = int(lines.pop(0))      # not sure what this number is
        list_of_outputs_lists = [self._data_element_tags, self._data_num_nodes_per_element, self._values]
        list_of_data_types = [int, int, float]
        list_of_positions = [0, 1, (2, self._entityDim, self._data_num_nodes_per_element)]
        self.__parse_lines(lines, list_of_outputs_lists, list_of_data_types, list_of_positions, self._number_of_elements_data)

    def __element_for_node(self):
        self._elem_list_for_nodes = []
        self._index_elem = []
        element_node_numbers = pd.DataFrame(self._element_node_numbers)
        for node in self._node_numbers:
            index = pd.Index
            i = -1
            while index.empty:
                i += 1
                index = element_node_numbers[element_node_numbers[i] == node].index
            self._elem_list_for_nodes.append(index[0])
            self._index_elem.append(i)

        # element_node_numbers = pd.DataFrame(self._element_node_numbers)
        # for node in self._node_numbers:
        #     column = []
        #     for i in range(len(element_node_numbers.loc[0, :])):
        #         index = element_node_numbers[element_node_numbers[i] == node].index
        #         column.append(index[0] if not index.empty else float('nan'))
        #     first_elem = np.nanmin(column)
        #     self._elem_list_for_nodes.append(int(first_elem))
        #     self._index_elem.append(column.index(first_elem))

        # for node in self._node_numbers:
        #     self._elem_list_for_nodes.append([i for i in range(len(self._element_node_numbers))
        #                                       if node in self._element_node_numbers[i]][0])
        #     self._index_elem.append(self._element_node_numbers[self._elem_list_for_nodes[-1]].index(node))

        # for node in self._node_numbers:
        #     n_idx = -1
        #     e_idx = 0
        #     while n_idx == -1:
        #         try:
        #             n_idx = self._element_node_numbers[e_idx].index(node)
        #             self._elem_list_for_nodes.append(e_idx)
        #             self._index_elem.append(n_idx)
        #         except ValueError:
        #             e_idx += 1

    def _node_values_dict(self):
        self.regions_dict = {}
        for reg_tag, elem_of_reg in zip(self._element_physical_tags, self._data_element_tags):
            if reg_tag not in self.regions_dict:
                self.regions_dict[reg_tag] = []
            self.regions_dict[reg_tag].append(elem_of_reg)

        self.elements_dict = {}
        for elem_tag, nodes_of_elem in zip(self._data_element_tags, self._element_node_numbers):
            self.elements_dict[elem_tag] = nodes_of_elem

        self.nodes_dict = {}
        for n_num, coors, elem_tag, idx in zip(self._node_numbers, self._node_coordinates,
                                               self._elem_list_for_nodes, self._index_elem):
            self.nodes_dict[n_num] = {}
            for i, key in zip(range(self._entityDim), ['x', 'y', 'z']):
                self.nodes_dict[n_num][key] = coors[i]
            for i, key in zip(range(self._entityDim), ['Vx', 'Vy', 'Vz']):
                self.nodes_dict[n_num][key] = self._values[elem_tag][idx * self._entityDim + i]
            self.nodes_dict[n_num]['e'] = self._data_element_tags[elem_tag]

    def _parse(self):
        """
        Call all parsing functions in the right sequence.
        """
        self._mesh_format()
        self._physical_names()
        self._nodes()
        self._elements_node_data()  # this one needs to be called before _elements method.
        self._elements()
        self.__element_for_node()
        self._node_values_dict()



from fiqus.pro_templates import combined
from jinja2 import Environment, FileSystemLoader


class ASS_PRO:
    def __init__(self, file_base_path, naming_conv=None):
        """
        Class for generating pro templates
        :param file_base_path: this is a full path to a folder with the model and includes model name, but without extension
        :param naming_conv: a dictionary with naming convention for regions in .pro file. Optional parameter, as a default below is used, but it could be overwritten if needed.
        """
        self.file_base_path = file_base_path
        if not naming_conv:
            self.naming_conv = {'omega': 'Omega', 'terms': 'Terms', 'bd': 'Bd', 'cond': '_c', 'powered': '_p', 'induced': '_i', 'air': '_a', 'line': 'Line'}
        else:
            self.naming_conv = naming_conv

    def assemble_combined_pro(self, template, rm, dm, BH_curves_path: str = ''):
        """
        Generates model .pro file from .pro template and regions model (rm)
        :param BH_curves_path: path of the BH curves pro file
        :param template: .pro template file name
        :param rm: regions model data structure (yaml loaded to regions data model)
        :param dm: data model structure
        :return: None. Generates .pro file and saves it on disk in the model folder under model_name.pro
        """
        loader = FileSystemLoader(combined.__path__)
        env = Environment(loader=loader, variable_start_string='<<', variable_end_string='>>',
                          trim_blocks=True, lstrip_blocks=True)
        env.globals.update(zip=zip)  # this is to pass python zip function to the template, as normally it is not available. It should work for passing any python function that is not available in .pro template.
        pro_template = env.get_template(template)
        output_from_parsed_template = pro_template.render(BHcurves=BH_curves_path, rm=rm, dm=dm, nc=self.naming_conv)
        with open(f"{self.file_base_path}.pro", "w") as tf:
            tf.write(output_from_parsed_template)

    def assemble_separate_pro(self, template, rm, dm, BH_curves_path: str = ''):
        """
        This function is not developed yet.
        """
        pass
